# Written and Refactor by keihan haqiq github@patpaws

from collections import OrderedDict
from typing import Generic, TypeVar
T = TypeVar('T')

class LRUCache(Generic[T]):
    """
    A Least Recently Used (LRU) cache implementation using an OrderedDict.

    Args:
        capacity (int): The maximum number of items to store in the cache.

    Attributes:
        _cache (OrderedDict): The underlying OrderedDict used for the cache.
        _capacity (int): The maximum number of items to store in the cache.

    Methods:
        get(key): Retrieve the value associated with the given key.
        put(key, value): Insert or update the value associated with the given key.
        monitor(): Print the keys in the cache in the order of least recently used to most recently used.
        __len__(): Return the number of items currently in the cache.
        __iter__() : Make it iterable
    """

    def __init__(self, capacity: int):
        self._cache = OrderedDict()
        self._capacity = capacity

    def get(self, key: T) -> T:
        """
        Retrieve the value associated with the given key.

        Args:
            key (T): The key to retrieve the value for.

        Returns:
            T: The value associated with the key, or None if the key is not found.
        """
        if key not in self._cache:
            return None

        # Move the requested item to the end (most recently used)
        self._cache.move_to_end(key)
        return self._cache[key]

    def put(self, key: T, value: T) -> None:
        """
        Insert or update the value associated with the given key.

        Args:
            key (T): The key to associate the value with.
            value (T): The value to associate with the key.
        """
        if key in self._cache:
            self._cache.move_to_end(key)
        self._cache[key] = value

        # If the cache exceeds its capacity, remove the least recently used item
        if len(self._cache) > self._capacity:
            self._cache.popitem(last=False)

    def monitor(self) -> None:
        """
        Print the keys in the cache in the order of least recently used to most recently used.
        """
        return [key for key in self._cache]

    def __len__(self) -> int:
        """
        Return the number of items currently in the cache.

        Returns:
            int: The number of items in the cache.
        """
        return len(self._cache)


    def __iter__(self) :
        """
        Return an iterator over the keys in the cache, in the order of least recently used to most recently used.

        Returns:
            Iterator[T]: An iterator over the keys in the cache.
        """
        return iter(self._cache)



# lru cache with 1000 records cawps.
def main() :
    cache = LRUCache(1000)
    cache.put("First item", 1)
    cache.put("Second item", 2)
    # Cache content: "Second item": 2, "First item": 1
    print(cache.get("First item"))  # 1
    # Cache content: "First item": 1, "Second item": 2
    cache.put("Third item", 3)
    # Cache content: "Third item": 3, "First item": 1
    print(cache.get("Second item"))  # -1 : not exists
    cache.put("Third item", 3)

    print('check len');
    print(len(cache));


# use for test case.
if __name__ == "__main__" :
    main();
